import * as WebSocket from "ws";
import { Server, IncomingMessage } from "http";
import { DatabaseManager } from "./database/database";
import Logging from "@hibas123/logging";
import Query from "./database/query";
import Session from "./database/session";
import nanoid = require("nanoid");

import * as JWT from "jsonwebtoken";

async function verifyJWT(token: string, publicKey: string) {
   return new Promise<any | undefined>((yes) => {
      JWT.verify(token, publicKey, (err, decoded) => {
         if (err)
            yes(undefined);
         else
            yes(decoded);
      })
   })
}

const StoreSym = Symbol("store");
function StoreQuery(result?: any) {
   return {
      [StoreSym]: true,
      result
   }
}

function DeleteQuery(result?: any) {
   return {
      [StoreSym]: false,
      result
   }
}

import { URLSearchParams } from "url";

type QueryTypes = "keys" | "get" | "set" | "update" | "delete" | "push" | "subscribe" | "unsubscribe";

export class ConnectionManager {
   static server: WebSocket.Server;

   static bind(server: Server) {
      this.server = new WebSocket.Server({ server });
      this.server.on("connection", this.onConnection.bind(this));
   }

   private static async onConnection(socket: WebSocket, req: IncomingMessage) {
      Logging.log("New Connection:");
      const sendError = (error: string) => socket.send(JSON.stringify({ ns: "error_msg", data: error }));


      const session = new Session(nanoid());

      let query = new URLSearchParams(req.url.split("?").pop());

      const database = query.get("database");
      const db = DatabaseManager.getDatabase(database);
      if (!db) {
         sendError("Invalid Database!");
         socket.close();
         return;
      }

      const accesskey = query.get("accesskey");
      if (db.accesskey) {
         if (!accesskey || accesskey !== db.accesskey) {
            sendError("Unauthorized!");
            socket.close();
            return;
         }
      }

      const authkey = query.get("authkey");
      if (authkey && db.publickey) {
         let res = await verifyJWT(authkey, db.publickey);
         if (!res || !res.uid) {
            sendError("Invalid JWT");
            socket.close();
            return;
         } else {
            session.uid = res.uid;
         }
      }

      const rootkey = query.get("rootkey");
      if (rootkey && db.rootkey) {
         if (rootkey === db.rootkey) {
            session.root = true;
            Logging.warning(`Somebody logged into ${database} via rootkey`);
         }
      }

      const stored = new Map<string, Query>();
      const answer = (id: string, data: any, error: boolean = false) => {
         socket.send(JSON.stringify({ ns: "message", data: { id, error, data } }));
      }

      const handler = new Map<string, ((data: any) => void)>();
      type QueryData = { id: string, type: QueryTypes, path: string[], data: any, options: any };

      const queryHandler = new Map<string, (query: Query, permissions: { read: boolean, write: boolean }, data?: any) => Promise<any>>();

      const noperm = new Error("No permisison!");

      if (session.root) {
         queryHandler.set("collections", (query, perm, data) => {
            return query.getCollections();
         })

         queryHandler.set("delete-collection", (query, perm, data) => {
            return query.deleteCollection();
         })
      }

      queryHandler.set("keys", (query, perm, data) => {
         if (!perm.read)
            throw noperm;
         return query.keys();
      })

      queryHandler.set("get", (query, perm, data) => {
         if (!perm.read)
            throw noperm;
         return query.get();
      })

      queryHandler.set("set", (query, perm, data) => {
         if (!perm.write)
            throw noperm;
         return query.set(data, {});
      })

      queryHandler.set("update", (query, perm, data) => {
         if (!perm.write)
            throw noperm;
         return query.update(data);
      })

      queryHandler.set("push", (query, perm, data) => {
         if (!perm.write)
            throw noperm;
         return query.push(data);
      })

      queryHandler.set("delete", (query, perm, data) => {
         if (!perm.write)
            throw noperm;
         return query.delete();
      })

      queryHandler.set("subscribe", async (query, perm, data) => {
         if (!perm.read)
            throw noperm;

         let subscriptionID = nanoid();

         query.subscribe((data) => {
            socket.send(JSON.stringify({ ns: "event", data: { id: subscriptionID, data } }));
         }, data.types, data.options);

         return StoreQuery(subscriptionID);
      })

      queryHandler.set("unsubscribe", async (query, perm, data) => {
         query.unsubscribe();
         return DeleteQuery(true);
      })

      handler.set("query", async ({ id, type, path, data }: QueryData) => {
         //TODO: Handle case with no id, type, path
         Logging.debug(`Request with id '${id}' from type '${type}' and path '${path}' with data`, data)

         try {
            if (!db)
               throw new Error("Database not found!");
            else {
               let handler = queryHandler.get(type);
               if (!handler) {
                  throw new Error("Invalid Request!");
               } else {
                  const query = stored.get(id) || db.getQuery(path || [], session.sessionid);
                  const perms = db.rules.hasPermission(path, session);

                  let res = await handler(query, perms, data);
                  if (res && typeof res === "object" && res[StoreSym] !== undefined) {
                     if (res[StoreSym])
                        stored.set(id, query);
                     else
                        stored.delete(id);
                  }
                  answer(id, res);
               }
            }
         } catch (err) {
            // Logging.error(err);
            Logging.debug("Sending error:", err);
            answer(id, err.message, true);
         }
      })

      socket.on("message", async (rawData: string) => {
         try {
            let message: { ns: string, data: any } = JSON.parse(rawData);
            let h = handler.get(message.ns);
            if (h) {
               h(message.data);
            }
         } catch (err) {
            Logging.errorMessage("Unknown Error:");
            Logging.error(err);
         }
      })

      socket.on("disconnect", () => {
         stored.forEach(query => query.unsubscribe());
         stored.clear();
         socket.removeAllListeners();
      })
   }
}